# Generated by Django 3.0.3 on 2020-04-01 14:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('avishan', '0009_auto_20200331_2041'),
    ]

    operations = [
        migrations.AddField(
            model_name='activity',
            name='data',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='activity',
            name='object_class',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='activity',
            name='object_id',
            field=models.BigIntegerField(blank=True, default=None, null=True),
        ),
    ]



import json
import sys

from django.contrib import messages
from django.core.handlers.wsgi import WSGIRequest
from django.http import JsonResponse
from django.utils import timezone

from avishan.configure import get_avishan_config
from avishan.exceptions import AvishanException, save_traceback


class Wrapper:
    """this middleware creates "current_request" storage for each incoming request"""

    def __init__(self, get_response):
        self.get_response = get_response
        get_avishan_config().on_startup()

    def __call__(self, request: WSGIRequest):
        from avishan.utils import discard_monitor, find_token, decode_token, add_token_to_response, find_and_check_user
        from avishan import current_request

        start_time = timezone.now()

        self.initialize_request_storage(current_request)
        current_request['request'] = request
        current_request['language'] = request.GET.get('language', current_request['language'])

        """Checks for avoid-touch requests"""
        if discard_monitor(current_request['request'].get_full_path()):
            print(f"NOT_MONITORED: {current_request['request'].get_full_path()}")
            return self.get_response(current_request['request'])

        current_request['start_time'] = start_time

        """Find token and parse it"""
        """
        Bara inke yadam nare. tooye sathe middleware vaghti error midim, chon nemidoonim api e ya template, error ro 
        zakhire mikonim mirim decorator, baad oonja k set shod in meghdar, check mikonim chon error hast, barmigarde 
        inja 
        """
        try:
            if find_token():
                decode_token()
                find_and_check_user()
        except AvishanException:
            pass
        except Exception as e:
            save_traceback()
            AvishanException(e)

        if current_request['language'] is None:
            current_request['language'] = get_avishan_config().LANGUAGE

        get_avishan_config().on_request()

        # todo 0.2.2 check for 'avishan_' in request bodies
        """Send request object to the next layer and wait for response"""
        try:
            response = self.get_response(current_request['request'])
            if not current_request['can_touch_response']:
                return response
        except AvishanException:
            pass
        except Exception as e:
            save_traceback()
            AvishanException(e)

        """messages"""
        if current_request['messages']['debug'] or current_request['messages']['info'] or \
                current_request['messages']['success'] or current_request['messages']['warning'] or \
                current_request['messages']['error']:
            # todo 0.2.3: check for debug=True

            if current_request['is_api']:
                current_request['response']['messages'] = current_request['messages']
            else:
                self.fill_messages_framework(current_request)
                # todo fix problem on template: not showing thrown exception message

        add_token_to_response(response)
        status_code = current_request['status_code']
        is_api = current_request['is_api']
        json_safe = not current_request['json_unsafe']
        if current_request['is_api']:
            response = current_request['response'].copy()

        if current_request['is_tracked'] or current_request['exception'] is not None:
            self.save_request_track(current_request)

        self.initialize_request_storage(current_request)

        if is_api:
            return JsonResponse(response, status=status_code, safe=json_safe)
        elif response.status_code == 200 and status_code != 200:
            response.status_code = status_code

        return response

    @staticmethod
    def initialize_request_storage(current_request):
        from avishan.models import RequestTrack
        current_request.clear()
        current_request['request'] = None
        current_request['response'] = {}
        current_request['can_touch_response'] = True
        current_request['is_tracked'] = False

        """If not checked "None", then switches between api & template"""
        current_request['is_api'] = None

        current_request['add_token'] = False
        current_request['view_name'] = "NOT_AVAILABLE"
        current_request['start_time'] = None
        current_request['end_time'] = None
        current_request['view_start_time'] = None
        current_request['view_end_time'] = None
        current_request['json_unsafe'] = False
        current_request['base_user'] = None
        current_request['user_group'] = None
        current_request['user_user_group'] = None
        current_request['authentication_object'] = None
        current_request['exception_record'] = None
        current_request['token'] = None
        current_request['decoded_token'] = None
        current_request['status_code'] = 200
        current_request['exception'] = None
        current_request['traceback'] = None
        current_request['language'] = None
        current_request['request_track_object'] = RequestTrack()
        current_request['context'] = {}
        current_request['messages'] = {
            'debug': [], 'info': [], 'success': [], 'warning': [], 'error': []
        }
        current_request['DEBUG'] = False

    @staticmethod
    def fill_messages_framework(current_request):
        for item in current_request['messages']['debug']:
            messages.debug(current_request['request'], item['body'])
        for item in current_request['messages']['info']:
            messages.info(current_request['request'], item['body'])
        for item in current_request['messages']['success']:
            messages.success(current_request['request'], item['body'])
        for item in current_request['messages']['warning']:
            messages.warning(current_request['request'], item['body'])
        for item in current_request['messages']['error']:
            messages.error(current_request['request'], item['body'])

    @staticmethod
    def save_request_track(current_request):
        from avishan.models import RequestTrackException
        for ignore in get_avishan_config().IGNORE_TRACKING_STARTS:
            if current_request['request'].get_full_path().startswith(ignore) and \
                    current_request['request_track_object']:
                current_request['request_track_object'].delete()
                return
        current_request['end_time'] = timezone.now()

        authentication_type_class_title = "NOT_AVAILABLE"
        authentication_type_object_id = 0
        if current_request['authentication_object']:
            authentication_type_class_title = current_request['authentication_object'].__class__.__name__
            authentication_type_object_id = current_request['authentication_object'].id

        request_data = "NOT_AVAILABLE"
        request_data_size = -1
        if current_request['request'].method in ['POST', 'PUT']:
            try:
                request_data = json.dumps(current_request['request'].data, indent=2)
                request_data_size = sys.getsizeof(json.dumps(current_request['request'].data))
            except:
                print("*DEBUG* request parse error")

        request_headers = ""
        for key, value in current_request['request'].META.items():
            if key.startswith('HTTP_'):
                request_headers += f'{key[5:]}={value}\n'
        for key in current_request['request'].FILES.keys():
            request_headers += f'FILE({key})\n'

        try:
            response_data = json.dumps(current_request['response'], indent=2)
        except:
            print("*DEBUG* response parse error:", current_request['response'])
            response_data = 'NOT_AVAILABLE'

        try:
            created = current_request['request_track_object'].update(
                view_name=current_request['view_name'],
                url=current_request['request'].get_full_path(),
                status_code=current_request['status_code'],
                method=current_request['request'].method,
                json_unsafe=current_request['json_unsafe'],
                is_api=current_request['is_api'],
                add_token=current_request['add_token'],
                user_user_group=current_request['user_user_group'],
                request_data=request_data,
                request_data_size=request_data_size,
                request_headers=request_headers,
                response_data=response_data,
                response_data_size=sys.getsizeof(response_data),
                start_time=current_request['start_time'],
                end_time=current_request['end_time'],
                total_execution_milliseconds=int((current_request['end_time'] - current_request[
                    'start_time']).total_seconds() * 1000),
                view_execution_milliseconds=int((current_request['view_end_time'] - current_request[
                    'view_start_time']).total_seconds() * 1000) if current_request['view_end_time'] else 0,
                authentication_type_class_title=authentication_type_class_title,
                authentication_type_object_id=authentication_type_object_id
            )

            if current_request['exception'] is not None:
                RequestTrackException.objects.create(
                    request_track=created,
                    class_title=current_request['exception'].__class__.__name__,
                    args=current_request['exception'].args,
                    traceback=current_request['traceback']
                )
        except Exception as e:
            print('save_request_track_error:'.upper(), e)



# Generated by Django 3.0.3 on 2020-02-15 22:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('avishan', '0004_usergroup_authenticate_with_phone_otp'),
    ]

    operations = [
        migrations.AddField(
            model_name='baseuser',
            name='language',
            field=models.CharField(default='EN', max_length=255),
        ),
    ]



import datetime
import inspect
import json
from typing import List, get_type_hints, Type, Callable, Optional, Union

from django.core.handlers.wsgi import WSGIRequest
from django.db import models
from django.db.models import QuerySet
from django.http import JsonResponse
from django.http.response import HttpResponseBase
from django.utils import timezone
from django.views import View

from avishan import current_request
from avishan.configure import get_avishan_config
from avishan.descriptor import DirectCallable, FunctionAttribute
from avishan.exceptions import ErrorMessageException, AvishanException, AuthException
from avishan.libraries.openapi3.classes import ApiDocumentation, Path, PathGetMethod, PathResponseGroup, \
    PathResponse, Content, Schema, PathPostMethod, PathRequest, PathPutMethod, PathDeleteMethod
from avishan.misc import status
from avishan.misc.translation import AvishanTranslatable
from avishan.models import AvishanModel, RequestTrack


# todo fix cors motherfucker
# todo doc
class AvishanView(View):
    authenticate: bool = True
    track_it: bool = False
    is_api: bool = None

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.response: dict = current_request['response']
        self.request: WSGIRequest
        self.class_attributes_type = get_type_hints(self.__class__)

        for kwarg_key, kwarg_value in kwargs.items():
            if hasattr(self, kwarg_key):
                self.__setattr__(kwarg_key, kwarg_value)

        for kwarg_key in request.GET.keys():
            kwarg_key: str
            if hasattr(self, kwarg_key):
                data = request.GET.getlist(kwarg_key)
                if len(data) > 1:
                    parsed = []
                    for item in data:
                        parsed.append(self.cast_data(kwarg_key, item))
                    self.__setattr__(kwarg_key, parsed)
                else:
                    self.__setattr__(kwarg_key, self.cast_data(kwarg_key, data[0]))

        self.current_request = current_request
        self.current_request['view_name'] = self.__class__.__name__
        self.current_request['request_track_exec'] = [
            {'title': 'begin', 'now': timezone.now()}
        ]
        self.current_request['is_api'] = self.is_api
        if self.track_it and not self.current_request['is_tracked']:
            self.current_request['is_tracked'] = True
            self.current_request['request_track_object'] = RequestTrack.objects.create()

    def http_method_not_allowed(self, request, *args, **kwargs):
        self.current_request['status_code'] = status.HTTP_405_METHOD_NOT_ALLOWED
        self.response['allowed_methods'] = self.get_allowed_methods()
        return super().http_method_not_allowed(request, *args, **kwargs)

    def get_allowed_methods(self) -> List[str]:
        allowed = []
        for method in ['get', 'post', 'put', 'delete']:
            if callable(getattr(self, method, False)):
                allowed.append(method)
        return allowed

    def dispatch(self, request, *args, **kwargs):
        if self.current_request['exception']:
            return

        try:
            if self.authenticate and not self.is_authenticated():
                raise AuthException(AuthException.TOKEN_NOT_FOUND)
            self.current_request['view_start_time'] = timezone.now()
            result = super().dispatch(request, *args, **kwargs)
            self.current_request['view_end_time'] = timezone.now()

        except AvishanException as e:
            raise e
        except Exception as e:
            AvishanException(wrap_exception=e)
            raise e
        if current_request['exception']:
            return
        return result

    def is_authenticated(self) -> bool:
        """
        Checks for user available in current_request storage
        :return: true if authenticated
        """
        if not self.current_request['authentication_object']:
            return False
        return True

    def cast_data(self, key, value):
        if key in self.class_attributes_type.keys():
            if isinstance(self.class_attributes_type[key], type):
                return self.class_attributes_type[key](value)
            return self.class_attributes_type[key].__args__[0](value)
        return value


class AvishanApiView(AvishanView):
    is_api = True
    authenticate = True

    def dispatch(self, request, *args, **kwargs):
        if self.current_request['request'].method not in ['GET', 'DELETE']:
            try:
                if len(current_request['request'].body) > 0:
                    current_request['request'].data = json.loads(current_request['request'].body.decode('utf-8'))
                else:
                    current_request['request'].data = {}
            except:
                current_request['request'].data = {}

        super().dispatch(request, *args, **kwargs)
        if current_request['can_touch_response']:
            return JsonResponse(self.response)
        return self.response


class AvishanTemplateView(AvishanView):
    is_api = False
    template_file_address: str = None
    template_url: str = None
    context: dict = {}

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.context = {
            **self.context,
            **{
                'CURRENT_REQUEST': self.current_request,
                'AVISHAN_CONFIG': get_avishan_config(),
                'self': self
            }
        }

    def dispatch(self, request, *args, **kwargs):
        self.parse_request_post_to_data()

        return super().dispatch(request, *args, **kwargs)

    def parse_request_post_to_data(self):
        if self.current_request['request'].method in ['POST', 'PUT']:
            try:
                if len(self.current_request['request'].body) > 0:
                    self.current_request['request'].data = dict(self.current_request['request'].POST)
                    for key, value in self.current_request['request'].data.items():
                        if len(value) == 1:
                            self.current_request['request'].data[key] = value[0]
                else:
                    self.current_request['request'].data = {}
            except:
                self.current_request['request'].data = {}

    def render(self):
        from django.shortcuts import render as django_render
        return django_render(self.request, self.template_file_address, self.context)


class AvishanModelApiView(AvishanApiView):
    track_it = True
    model: Type[AvishanModel] = None
    model_item: AvishanModel = None
    model_function: Callable = None  # todo these sends model not dict
    model_function_name: str = None
    direct_callable: Optional[DirectCallable] = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        model_plural_name = kwargs.get('model_plural_name', None)
        model_item_id = kwargs.get('model_item_id', None)
        self.model_function_name = kwargs.get('model_function_name', None)
        self.model: Union[AvishanModel, type] = AvishanModel.get_model_by_plural_snake_case_name(model_plural_name)
        if not self.model:
            raise ErrorMessageException('Entered model name not found')

        if self.model_function_name is None:
            if request.method == 'GET':
                if model_item_id is None:
                    self.model_function_name = 'all'
                else:
                    self.model_function_name = 'get'
            elif request.method == 'POST':
                self.model_function_name = 'create'
            elif request.method == 'PUT':
                self.model_function_name = 'update'
            elif request.method == 'DELETE':
                self.model_function_name = 'remove'

        if model_item_id is not None:
            self.model_item = self.model.get(avishan_raise_400=True, id=int(model_item_id))
        if self.model_function_name is not None:
            for direct_callable in self.model.direct_callable_methods():
                if self.model_function_name == direct_callable.name:
                    self.direct_callable = direct_callable
                    break
            if not self.direct_callable:
                raise AuthException(AuthException.METHOD_NOT_DIRECT_CALLABLE)
            self.authenticate = self.direct_callable.authenticate
            try:
                if self.model_item is None:
                    self.model_function = getattr(self.model, self.model_function_name)
                else:
                    self.model_function = getattr(self.model_item, self.model_function_name)
            except AttributeError:
                raise ErrorMessageException(AvishanTranslatable(
                    EN=f'Requested method not found in model {self.model.class_name()}'
                ))

    def get(self, request, *args, **kwargs):
        if self.model_function_name == 'get':
            result = self.model_item
        else:
            result = self.model_function()

            if isinstance(result, QuerySet):
                result = self.model.queryset_handler(request.GET, queryset=result)
        response = self.parse_returned_data(result)
        if current_request['can_touch_response']:
            self.response[self.direct_callable.response_json_key] = response
        else:
            self.response = response

    def post(self, request, *args, **kwargs):
        if self.direct_callable.dismiss_request_json_key:
            data = request.data
        else:
            data = request.data[self.direct_callable.request_json_key]

        data = self.parse_request_data(**data)
        result = self.model_function(**data)
        response = self.parse_returned_data(result)
        if current_request['can_touch_response']:
            self.response[self.direct_callable.response_json_key] = response
        else:
            self.response = response

    def put(self, request, *args, **kwargs):
        self.post(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        result = self.model_function()
        response = self.parse_returned_data(result)
        if current_request['can_touch_response']:
            self.response[self.direct_callable.response_json_key] = response
        else:
            self.response = response

    @staticmethod
    def parse_returned_data(returned):
        if isinstance(returned, HttpResponseBase):
            current_request['can_touch_response'] = False
            return returned
        if isinstance(returned, QuerySet):
            return [item.to_dict() for item in returned]
        elif isinstance(returned, list):
            if len(returned) > 0 and isinstance(returned[0], AvishanModel):
                return [item.to_dict() for item in returned]
            return returned
        elif isinstance(returned, AvishanModel):
            return returned.to_dict()
        else:
            return returned

    def parse_request_data(self, **kwargs):
        """Parse Request Body

        Request body consists of multiple items, some have primitive data, like name:str or price:float. But for
        relational objects like birth_place:City, we should convert json-acceptable data to specific python objects. In
        this case we should get corresponding object from database.
        """

        cleaned = {}
        for function_attribute in self.direct_callable.args:
            function_attribute: FunctionAttribute

            """Errors for required data before python raises exception for function params"""
            if function_attribute.is_required and function_attribute.name not in kwargs.keys():
                raise ErrorMessageException(f'field "{function_attribute.name}" is required')

            """Optional field not provided"""
            if function_attribute.name not in kwargs.keys():
                continue

            """Objects and array of objects should convert to db objects"""
            if function_attribute.type is function_attribute.TYPE.OBJECT:
                cleaned[function_attribute.name] = self._parse_request_object(
                    function_attribute=function_attribute, target_dict=kwargs[function_attribute.name]
                )
            elif function_attribute.type is function_attribute.TYPE.ARRAY and \
                    inspect.isclass(function_attribute.type_of) and \
                    issubclass(function_attribute.type_of, AvishanModel):
                cleaned[function_attribute.name] = [
                    self._parse_request_object(
                        function_attribute=function_attribute, target_dict=item
                    ) for item in kwargs[function_attribute.name]
                ]
            elif function_attribute.type is function_attribute.TYPE.FILE:
                raise NotImplementedError()
            else:
                """Casts primitive data here"""
                cleaned[function_attribute.name] = function_attribute.type_caster(
                    entry=kwargs[function_attribute.name],
                    target_type=function_attribute.type
                )

        return cleaned

    @staticmethod
    def _parse_request_object(function_attribute: FunctionAttribute, target_dict: dict):
        """Parse object

        Get objects from database

        :param FunctionAttribute function_attribute: corresponding attribute data
        :param dict target_dict: request data
        :return: cleaned object
        """

        if not isinstance(target_dict, dict):
            raise ErrorMessageException(f'value for "{function_attribute.name}" must be dict consist of id or other '
                                        f'unique values so that db can find corresponding object')

        return function_attribute.type_of.get_from_dict(target_dict)


class PasswordHash(AvishanApiView):
    authenticate = False

    def get(self, request, *args, **kwargs):
        import bcrypt
        current_request['response']['hashed'] = bcrypt.hashpw(kwargs['password'].encode('utf8'),
                                                              bcrypt.gensalt()).decode('utf8')
        return JsonResponse(current_request['response'])



import datetime
import json

from django.http import JsonResponse
from django.utils import timezone

from avishan.exceptions import AvishanException, AuthException
from . import current_request
from .models import RequestTrack


class AvishanViewDecorator:
    def __init__(self, is_api: bool, methods=('GET',), authenticate: bool = None, track_it: bool = False):
        self.methods = methods
        self.authenticate = authenticate
        self.is_api = is_api
        self.track_it = track_it

    def __call__(self, view_function):

        def wrapper(*args, **kwargs):
            current_request['view_name'] = view_function.__name__
            current_request['request_track_exec'] = [
                {'title': 'begin', 'now': timezone.now()}
            ]
            current_request['is_api'] = self.is_api
            if self.track_it and not current_request['is_tracked']:
                current_request['is_tracked'] = True
                current_request['request_track_object'] = RequestTrack.objects.create()
            if current_request['exception']:
                """If we have exception here, should return after "is_api" assignment to middleware"""
                return JsonResponse({})

            try:
                """
                If user not provided, return with error.
                """
                self.before_request()

                if self.authenticate and not self.is_authenticated():
                    raise AuthException(AuthException.ACCESS_DENIED)

                """http method check and raise 405"""
                if current_request['is_api'] and current_request['request'].method not in self.methods:
                    raise AuthException(AuthException.HTTP_METHOD_NOT_ALLOWED)

                self.after_request()

                current_request['view_start_time'] = timezone.now()
                result = view_function(*args, **kwargs)
                current_request['view_end_time'] = timezone.now()

                self.before_response()

                self.after_response()

            except AvishanException:
                return JsonResponse({})
            except Exception as e:
                AvishanException(wrap_exception=e)
                return JsonResponse({})
            if current_request['exception']:
                return JsonResponse({})

            return result

        return wrapper

    def before_request(self):
        raise NotImplementedError()

    def after_request(self):
        raise NotImplementedError()

    def before_response(self):
        raise NotImplementedError()

    def after_response(self):
        raise NotImplementedError()

    @staticmethod
    def is_authenticated() -> bool:
        """
        Checks for user available in current_request storage
        :return: true if authenticated
        """
        from . import current_request
        if not current_request['authentication_object']:
            return False
        return True


class AvishanApiViewDecorator(AvishanViewDecorator):

    def __init__(self, methods=('GET',), authenticate: bool = True, track_it: bool = False):
        super().__init__(is_api=True, methods=methods, authenticate=authenticate, track_it=track_it)

    def before_request(self):
        if current_request['request'].method not in ['GET', 'DELETE']:
            try:
                if len(current_request['request'].body) > 0:
                    current_request['request'].data = json.loads(current_request['request'].body.decode('utf-8'))
                else:
                    current_request['request'].data = {}
            except:
                current_request['request'].data = {}

    def after_request(self):
        pass

    def before_response(self):
        pass

    def after_response(self):
        pass


class AvishanTemplateViewDecorator(AvishanViewDecorator):
    def __init__(self, methods=('GET',), authenticate: bool = True, track_it: bool = False):
        super().__init__(is_api=False, methods=methods, authenticate=authenticate, track_it=track_it)

    def before_request(self):
        if current_request['request'].method in ['POST', 'PUT']:
            try:
                if len(current_request['request'].body) > 0:
                    current_request['request'].data = dict(current_request['request'].POST)
                    for key, value in current_request['request'].data.items():
                        if len(value) == 1:
                            current_request['request'].data[key] = value[0]
                else:
                    current_request['request'].data = {}
            except:
                current_request['request'].data = {}

    def after_request(self):
        pass

    def before_response(self):
        pass

    def after_response(self):
        pass



from typing import List, Tuple

import requests

from avishan.configure import get_avishan_config


def distance_matrix(origins: List[Tuple[float, float]], destinations: List[Tuple[float, float]]) -> \
        List[List[Tuple[float, float]]]:
    if len(origins) == 0 or len(destinations) == 0:
        raise ValueError('length of entered lists can\'t be 0')
    origins_text = ''
    for item in origins:
        origins_text += f'|{item[0]},{item[1]}'
    destinations_text = ''
    for item in destinations:
        destinations_text += f'|{item[0]},{item[1]}'

    url = f'https://api.neshan.org/v1/distance-matrix?' \
          f'origins={origins_text[1:]}&destinations={destinations_text[1:]}'
    response = requests.get(
        url=url,
        headers={'Api-Key': get_avishan_config().NESHAN_API_KEY}
    ).json()

    data = []
    for row in response['rows']:
        data.append([(element['distance']['value'], element['duration']['value']) for element in row['elements']])
    return data



from typing import Tuple, List, Optional, Union

import yaml
from django.db.models.base import ModelBase
from django.utils import timezone

from avishan.configure import get_avishan_config
from avishan.descriptor import Attribute, ApiMethod, Model, DjangoModel, DirectCallable, FunctionAttribute


# todo tags
# todo POST body? 
class OpenApi:

    def __init__(self,
                 application_title: str,
                 application_description: str = None,
                 application_version: str = '0.1.0',
                 application_servers: List['Server'] = None
                 ):
        self.application_title = application_title
        self.application_description = application_description
        self.application_version = application_version
        self.application_servers = application_servers
        self.models: List[Model] = sorted(get_avishan_config().get_openapi_schema_models(), key=lambda x: x.name)

    def export(self) -> dict:
        # todo security
        data = {
            'openapi': '3.0.1',
            'info': self._export_info(),
            'servers': self._export_servers(),
            'tags': self._export_tags(),
            'paths': self._export_paths(),
            'components': self._export_components()
        }
        return data

    def _export_info(self):
        return {
            'title': self.application_title,
            'description': self.application_description,
            'version': self.application_version
        }

    def _export_servers(self):
        return [
            {
                'url': item.url,
                'description': item.description
            }
            for item in self.application_servers
        ]

    def _export_tags(self) -> list:
        # todo https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#tagObject
        return []

    def _export_paths(self) -> dict:
        data = {}
        for model in self.models:
            if model.name in get_avishan_config().get_openapi_ignored_path_models():
                continue
            for api_method in model.methods:
                api_method: DirectCallable
                if api_method.hide_in_redoc:
                    continue
                if api_method.url not in data.keys():
                    data[api_method.url] = Path(url=api_method.url)
                setattr(data[api_method.url], api_method.method.name.lower(), Operation(
                    summary=api_method.short_description,
                    description=api_method.long_description,
                    request_body=Operation.extract_request_body_from_api_method(api_method),
                    responses=Operation.extract_responses_from_api_method(api_method),
                ))

        for key, value in data.items():
            data[key] = value.export()

        return data

    def _export_components(self) -> dict:
        data = {
            'schemas': self._export_schemas(),
            'responses': self._export_responses(),
            'parameters': self._export_parameters(),
            'examples': self._export_examples(),
            'requestBodies': self._export_request_bodies(),
            'headers': self._export_headers(),
            'securitySchemes': self._export_securitySchemes(),
        }
        delete_list = []
        for key, value in data.items():
            if value == {}:
                delete_list.append(key)
        for item in delete_list:
            del data[item]
        return data

    def _export_schemas(self) -> dict:
        schemas = {}
        for model in self.models:
            schemas[model.name] = Schema.create_from_model(model).export()
        return schemas

    def _export_responses(self) -> dict:
        # todo
        return {}

    def _export_parameters(self) -> dict:
        # todo
        return {}

    def _export_examples(self) -> dict:
        # todo
        return {}

    def _export_request_bodies(self) -> dict:
        # todo
        return {}

    def _export_headers(self) -> dict:
        # todo
        return {}

    def _export_securitySchemes(self) -> dict:
        # todo
        return {}

    def export_yaml(self) -> str:
        return yaml.dump(self.export())


class Server:
    def __init__(self, url: str, description: str = None):
        self.url = url
        self.description = description


class Tag:

    def __init__(self, name: str):
        self.name = name


class Parameter:
    # todo
    pass


class Property:

    def __init__(self, name: str, schema: 'Schema', required: bool = True, default=Attribute.NO_DEFAULT):
        self.name = name
        self.schema = schema
        self.required = required
        self.default = default

    @classmethod
    def create_from_attribute(cls, attribute: Attribute, request_body_related: bool = False) -> 'Property':
        required = True
        if isinstance(attribute, FunctionAttribute):
            required = attribute.is_required
        created = Property(
            name=attribute.name,
            schema=Schema.create_from_attribute(attribute, request_body_related),
            required=required,
            default=attribute.default
        )
        created.schema.default = created.default
        return created

    def export(self) -> dict:
        return self.schema.export()


class Schema:
    _TYPE_POOL = {
        Attribute.TYPE.STRING: ('string', None),
        Attribute.TYPE.INT: ('integer', None),
        Attribute.TYPE.FLOAT: ('number', 'float'),
        Attribute.TYPE.DATE: ('string', 'date'),
        Attribute.TYPE.TIME: ('string', 'time'),
        Attribute.TYPE.DATETIME: ('string', 'date-time'),
        Attribute.TYPE.BOOLEAN: ('boolean', None),
        Attribute.TYPE.ARRAY: ('array', None)
    }

    def __init__(self,
                 name: str = None,
                 type: str = None,
                 format: str = None,
                 default=Attribute.NO_DEFAULT,
                 items: 'Schema' = None,
                 properties: List[Property] = None,
                 description: str = None
                 ):
        if properties is None:
            properties = []

        self.name = name
        self.type = type
        self.format = format
        self.default = default
        self.items = items
        self.properties = properties
        self.description = description

    @classmethod
    def create_from_attribute(cls, attribute: Attribute, request_body_related: bool = False) -> 'Schema':
        if attribute.type is Attribute.TYPE.OBJECT:
            if request_body_related:
                return Schema(
                    type='object',
                    properties=[
                        Property(name='id', schema=Schema(type='integer'))
                    ]
                )
            else:
                return Schema(name=attribute.type_of.__name__)
        if attribute.type is Attribute.TYPE.FILE:
            return Schema(name='File')

        create_kwargs = {
            'type': cls.type_exchange(attribute.type)[0],
            'format': cls.type_exchange(attribute.type)[1],
            'description': attribute.description
        }
        if attribute.type is Attribute.TYPE.ARRAY:
            create_kwargs['items'] = cls.type_exchange(attribute.type_of)
            if isinstance(create_kwargs['items'], tuple):
                create_kwargs['items'] = Schema(type=create_kwargs['items'][0])

        return Schema(**create_kwargs)

    @classmethod
    def create_object_from_args(cls, args: List[Attribute], request_body_related: bool = False) -> 'Schema':
        return Schema(
            type='object',
            properties=[Property.create_from_attribute(item, request_body_related) for item in args]
        )

    @classmethod
    def create_from_model(cls, model: DjangoModel) -> 'Schema':
        return cls.create_object_from_args(model.attributes)

    @classmethod
    def type_exchange(cls, entry: Union[Attribute.TYPE, ModelBase]) -> Union[Tuple[str, str], 'Schema']:
        try:
            return cls._TYPE_POOL[entry]
        except KeyError:
            if isinstance(entry, ModelBase):
                return Schema(name=entry.__name__)

        raise NotImplementedError()

    def export(self) -> dict:
        if self.name:
            return {
                "$ref": f"#/components/schemas/{self.name}"
            }
        data = {
            'type': self.type
        }
        if self.format:
            data['format'] = self.format
        if self.default is not Attribute.NO_DEFAULT:
            data['default'] = self.default
        if self.description:
            data['description'] = self.description
        if self.items:
            data['items'] = self.items.export()

        if len(self.properties) > 0:
            data['properties'] = {}
            data['required'] = []
            for item in self.properties:
                if item.required:
                    data['required'].append(item.name)
                data['properties'][item.name] = item.export()
            if len(data['required']) == 0:
                del data['required']

        if self.type == 'string' and self.format == 'date-time':
            data['example'] = timezone.now().strftime(get_avishan_config().DATETIME_STRING_FORMAT)
        if self.type == 'string' and self.format == 'date':
            data['example'] = timezone.now().strftime(get_avishan_config().DATE_STRING_FORMAT)

        return data


class Content:
    def __init__(self, schema: Schema, examples=None):
        self.schema = schema
        self.examples = examples  # todo

    @classmethod
    def create_from_attribute(cls, attribute: Attribute) -> 'Content':
        return Content(
            Schema.create_from_attribute(attribute)
        )

    def export(self) -> dict:
        data = {
            "application/json": {
                "schema": self.schema.export()
            }
        }
        if self.examples:
            data['application/json']['examples'] = self.examples
        return data


class RequestBody:
    def __init__(self, content: Content, required: bool = True):
        self.content = content
        self.required = required

    def export(self) -> dict:
        return {
            'content': self.content.export(),
            'required': self.required
        }


class Response:
    def __init__(self, status_code: int, description: str = None, content: Content = None):
        self.status_code = status_code
        self.description = description
        self.content = content

    def export(self) -> dict:
        data = {}
        if self.description:
            data['description'] = self.description
        if self.content:
            data['content'] = self.content.export()
        return data


class Operation:
    def __init__(self,
                 tags: List[Tag] = None,
                 summary: str = None,
                 description: str = None,
                 parameters: List[Parameter] = None,
                 request_body: RequestBody = None,
                 responses: List[Response] = None
                 ):
        if tags is None:
            tags = []
        if parameters is None:
            parameters = []
        if responses is None:
            responses = []

        self.tags = tags
        self.summary = summary
        self.description = description
        self.parameters = parameters
        self.request_body = request_body
        self.responses = responses

    @staticmethod
    def extract_request_body_from_api_method(api_method: DirectCallable) -> Optional[RequestBody]:
        if len(api_method.args) == 0:
            return None
        content = Content(schema=Schema.create_object_from_args(api_method.args, request_body_related=True))
        if not api_method.dismiss_request_json_key:
            content.schema = Schema(
                type='object',
                properties=[Property(
                    name=api_method.request_json_key,
                    schema=content.schema
                )]
            )
        return RequestBody(content=content)

    @staticmethod
    def extract_responses_from_api_method(api_method: DirectCallable) -> List[Response]:
        responses = []
        for item in api_method.responses:
            content = Content.create_from_attribute(item.returns)
            if not api_method.dismiss_response_json_key:
                content.schema = Schema(
                    type='object',
                    properties=[Property(
                        name=api_method.response_json_key,
                        schema=content.schema
                    )]
                )
            responses.append(Response(
                status_code=item.status_code,
                description=item.description,
                content=content
            ))

        return responses

    def export(self) -> dict:
        data = {}
        if len(self.tags) > 0:
            data['tags'] = [item for item in self.tags]
        if self.summary:
            data['summary'] = self.summary
        if self.description:
            data['description'] = self.description
        if len(self.parameters) > 0:
            raise NotImplementedError()
        if self.request_body:
            data['requestBody'] = self.request_body.export()
        if len(self.responses) > 0:
            data['responses'] = {}
            for item in self.responses:
                data['responses'][str(item.status_code)] = item.export()

        return data


class Path:
    def __init__(self,
                 url: str,
                 summary: str = None,
                 description: str = None,
                 get: Operation = None,
                 post: Operation = None,
                 put: Operation = None,
                 delete: Operation = None,
                 ):
        self.url = url
        self.summary = summary
        self.description = description
        self.get = get
        self.post = post
        self.put = put
        self.delete = delete

    def export(self) -> dict:
        data = {}
        if self.get:
            data['get'] = self.get.export()
        if self.post:
            data['post'] = self.post.export()
        if self.put:
            data['put'] = self.put.export()
        if self.delete:
            data['delete'] = self.delete.export()
        if self.description:
            data['description'] = self.description
        if self.summary:
            data['summary'] = self.summary

        return data



from django.db import models

from avishan.models import *


class AuthenticationType(AvishanModel):
    user_user_group = models.OneToOneField(UserUserGroup, on_delete=models.CASCADE, help_text='Target UserUserGroup')
    last_used = models.DateTimeField(default=None, blank=True, null=True, help_text='Last time user sent request')
    last_login = models.DateTimeField(default=None, blank=True, null=True, help_text='Last time user logged in')
    last_logout = models.DateTimeField(default=None, blank=True, null=True, help_text='Last time user logged out')

    export_ignore = True

    django_admin_raw_id_fields = [user_user_group]
    django_admin_list_display = ['key', user_user_group, last_used, last_login, last_logout]
    django_admin_list_filter = [user_user_group]
    django_admin_search_fields = ['key']

    class Meta:
        abstract = True

    @classmethod
    def direct_callable_methods(cls):
        return super().direct_callable_methods() + [
            DirectCallable(
                model=cls,
                target_name='login',
                response_json_key=cls.class_snake_case_name(),
                method=DirectCallable.METHOD.POST,
                authenticate=False
            )
        ]

    @classmethod
    def _register(cls, user_user_group: UserUserGroup, key: str, **kwargs) -> 'AuthenticationType':
        from avishan.exceptions import AuthException

        try:
            key_item = cls.key_field().related_model.get(key=key)
        except cls.key_field().related_model.DoesNotExist:
            key_item = cls.key_field().related_model.create(key)
        try:
            cls.objects.get(**{
                'key': key_item,
                'user_user_group__user_group': user_user_group.user_group
            })
            raise AuthException(AuthException.DUPLICATE_AUTHENTICATION_IDENTIFIER)
        except cls.DoesNotExist:
            pass

        creation_dict = {
            **{
                'user_user_group': user_user_group,
                'key': key_item,
            },
            **kwargs
        }

        return cls.create(**creation_dict)

    @classmethod
    def login(cls, key: str, user_group_title: str = None, **kwargs) -> 'AuthenticationType':
        from avishan.exceptions import AuthException

        try:
            if not user_group_title:
                found_object: KeyValueAuthentication = cls.objects.get(
                    key=cls.key_field().related_model.get(key=key))
            else:
                found_object: KeyValueAuthentication = cls.objects.get(**{
                    'key': cls.key_field().related_model.get(key=key),
                    'user_user_group__user_group__title': user_group_title
                })

        except cls.DoesNotExist:
            raise AuthException(AuthException.ACCOUNT_NOT_FOUND)
        except cls.MultipleObjectsReturned:
            raise AuthException(AuthException.MULTIPLE_CONNECTED_ACCOUNTS)
        kwargs['found_object'] = found_object
        kwargs['submit_login'] = True

        cls._login_post_check(kwargs)

        if kwargs['submit_login']:
            found_object._submit_login()
        return found_object

    @classmethod
    def _login_post_check(cls, kwargs):
        """
        Checks for post login
        :param kwargs:
        :type kwargs:
        """
        pass

    def _submit_login(self):
        self.last_login = timezone.now()
        self.last_used = None
        self.last_logout = None
        self.save()
        self.populate_current_request()

    def _submit_logout(self):
        self.last_logout = timezone.now()
        self.save()
        current_request['authentication_object'] = None
        current_request['add_token'] = False

    def populate_current_request(self):
        current_request['base_user'] = self.user_user_group.base_user
        current_request['user_group'] = self.user_user_group.user_group
        current_request['user_user_group'] = self.user_user_group
        current_request['authentication_object'] = self
        if current_request['language'] is None:
            current_request['language'] = self.user_user_group.base_user.language
        current_request['add_token'] = True

    @classmethod
    def key_field(cls) -> models.ForeignKey:
        raise NotImplementedError()


class KeyValueAuthentication(AuthenticationType):
    hashed_password = models.CharField(max_length=255, blank=True, null=True, default=None, help_text='Hashed password')

    to_dict_private_fields = [hashed_password]

    class Meta:
        abstract = True

    @classmethod
    def register(cls, user_user_group: UserUserGroup, key: str, password: Optional[str] = None) -> \
            Union['EmailPasswordAuthenticate', 'PhonePasswordAuthenticate']:

        data = {
            'user_user_group': user_user_group,
            'key': key
        }

        if password is not None:
            data['hashed_password'] = cls._hash_password(password)

        return cls._register(**data)

    # noinspection PyMethodOverriding
    @classmethod
    def login(cls, key: str, password: str, user_group_title: str = None, **kwargs) -> 'KeyValueAuthentication':
        return super().login(key=key, user_group_title=user_group_title, password=password, **kwargs)

    def add_password(self, password: str) -> bool:
        if self.hashed_password is None:
            self.hashed_password = self._hash_password(password)
            self.save()
            return True
        return False

    @classmethod
    def key_field(cls) -> Union[models.ForeignKey, models.Field]:
        return cls.get_field('key')

    @staticmethod
    def _hash_password(password: str) -> str:
        """
        Hash entered password
        :param password:
        :return: hashed password in string
        """
        import bcrypt
        return bcrypt.hashpw(password.encode('utf8'), bcrypt.gensalt()).decode('utf8')

    @staticmethod
    def _check_password(password: str, hashed_password: str) -> bool:
        """
        compares password with hashed instance.
        :param password:
        :param hashed_password:
        :return: True if match
        """
        import bcrypt
        return bcrypt.checkpw(password.encode('utf8'), hashed_password.encode('utf8'))

    @classmethod
    def _login_post_check(cls, kwargs):
        from avishan.exceptions import AuthException

        if not cls._check_password(kwargs['password'], kwargs['found_object'].hashed_password):
            raise AuthException(AuthException.INCORRECT_PASSWORD)


class EmailPasswordAuthenticate(KeyValueAuthentication):
    key = models.ForeignKey(Email, on_delete=models.CASCADE, related_name='password_authenticates',
                            help_text='Related Email object')


class PhonePasswordAuthenticate(KeyValueAuthentication):
    key = models.ForeignKey(Phone, on_delete=models.CASCADE, related_name='password_authenticates',
                            help_text='Related Phone object')


class OtpAuthentication(AuthenticationType):
    code = models.CharField(max_length=255, blank=True, null=True, help_text='Code sent to user')
    date_sent = models.DateTimeField(null=True, blank=True, default=None, help_text='Date code sent to user')
    tried_codes = models.TextField(blank=True, default="", help_text='Incorrect user tried codes')

    to_dict_private_fields = [code, tried_codes, date_sent]

    class Meta:
        abstract = True

    @classmethod
    def register(cls, user_user_group: UserUserGroup, key: str) -> 'OtpAuthentication':

        data = {
            'user_user_group': user_user_group,
            'key': key
        }
        return cls._register(**data)

    @classmethod
    def login(cls, key: str, user_group_title: str = None, **kwargs) -> 'OtpAuthentication':
        return super().login(key=key, user_group_title=user_group_title, verify=False)

    @classmethod
    def _login_post_check(cls, kwargs):
        from avishan.exceptions import ErrorMessageException, AuthException

        found_object: PhoneOtpAuthenticate = kwargs['found_object']
        if isinstance(found_object, PhoneOtpAuthenticate):
            gap = get_avishan_config().POA_VERIFICATION_GAP_SECONDS
        else:
            raise NotImplementedError()

        if not kwargs.get('verify', False):
            kwargs['submit_login'] = False
            if found_object.date_sent and (timezone.now() - found_object.date_sent).total_seconds() < gap:
                raise ErrorMessageException(AvishanTranslatable(
                    EN='Verification code sent recently, Please try again later',
                    FA='برای ارسال مجدد کد، کمی صبر کنید'
                ))
            if get_avishan_config().ASYNC_AVAILABLE:
                from avishan.tasks import async_phone_otp_authentication_send_otp_code
                async_phone_otp_authentication_send_otp_code.delay(poa_id=found_object.id)
            else:
                found_object.send_otp_code()

        else:
            code = kwargs['entered_code']
            if not found_object._check_entered_code(code):
                raise AuthException(AuthException.INCORRECT_PASSWORD)
            if found_object.last_login is None:
                current_request['status_code'] = status.HTTP_201_CREATED
            found_object.code = None
            found_object.date_sent = None
            found_object.tried_codes = ""
            found_object.save()
            if found_object.key.date_verified is None:
                found_object.key.date_verified = timezone.now()
                found_object.key.save()

    @classmethod
    def verify(cls, key: str, entered_code: str, user_group: UserGroup = None) -> 'OtpAuthentication':

        try:
            cls.key_field().related_model.get(key=key)
        except cls.key_field().related_model.DoesNotExist:
            cls.key_field().related_model.create(key)

        return super().login(
            key=key,
            user_group=user_group,
            verify=True,
            entered_code=entered_code
        )

    def send_otp_code(self):
        self.code = self.create_otp_code()
        self.date_sent = timezone.now()
        self.tried_codes = ""
        self.save()

    def _check_entered_code(self, entered_code: str) -> bool:
        from avishan.exceptions import ErrorMessageException

        if isinstance(self, PhoneOtpAuthenticate):
            valid_seconds = get_avishan_config().POA_VERIFICATION_VALID_SECONDS
        else:
            raise NotImplementedError()

        if self.code is None:
            raise ErrorMessageException(AvishanTranslatable(
                EN='Code not found for this account',
                FA='برای این حساب کدی پیدا نشد'
            ))
        if (timezone.now() - self.date_sent).total_seconds() > valid_seconds:
            self.code = None
            self.save()
            raise ErrorMessageException(AvishanTranslatable(
                EN='Code Expired',
                FA='کد منقضی شده است'
            ))

        if self.code != entered_code:
            self.tried_codes += f"{entered_code}\n"
            self.save()
            return False

        return True

    @classmethod
    def create_otp_code(cls) -> str:
        if cls._meta.model is PhoneOtpAuthenticate:
            length = get_avishan_config().POA_VERIFICATION_CODE_LENGTH
        else:
            raise NotImplementedError()

        return str(random.randint(10 ** (length - 1), 10 ** length - 1))

    @classmethod
    def key_field(cls) -> Union[models.Field, models.ForeignKey]:
        return cls.get_field('key')


class PhoneOtpAuthenticate(OtpAuthentication):
    key = models.ForeignKey(Phone, on_delete=models.CASCADE, related_name='otp_authenticates',
                            help_text='Related Phone object')

    def send_otp_code(self):
        from avishan.exceptions import ErrorMessageException
        super().send_otp_code()

        if get_avishan_config().KAVENEGAR_SMS_ENABLE:
            self.key.send_sms(
                template=get_avishan_config().KAVENEGAR_DEFAULT_TEMPLATE,
                token=self.code
            )
        else:
            raise ErrorMessageException(AvishanTranslatable(
                EN='SMS Provider not found. Enable in "SMS Providers" avishan config section'
            ))


class VisitorKey(AuthenticationType):
    key = models.CharField(max_length=255, unique=True, help_text='Random generated key')

    django_admin_list_display = key,
    django_admin_search_fields = key,

    @classmethod
    def create(cls, key: str) -> 'VisitorKey':
        return super().create(key=key)

    @classmethod
    def key_field(cls) -> models.Field:
        return cls.get_field('key')

    @staticmethod
    def create_key() -> str:
        import secrets
        return secrets.token_urlsafe(get_avishan_config().VISITOR_KEY_LENGTH)

    @classmethod
    def register(cls, user_user_group: UserUserGroup) -> 'VisitorKey':

        key = cls.create_key()
        while True:
            try:
                cls.get(key=key)
                key = cls.create_key()
            except cls.DoesNotExist:
                break

        data = {
            'user_user_group': user_user_group,
            'key': key,
        }

        return cls.objects.create(**data)

    @classmethod
    def login(cls, key: str, user_group_title: str = None) -> 'VisitorKey':
        from avishan.exceptions import AuthException

        try:
            found_object = cls.objects.get(
                **{
                    'key': key,
                    "user_user_group__user_group__title": user_group_title
                }
            )
        except cls.DoesNotExist:
            raise AuthException(AuthException.ACCOUNT_NOT_FOUND)

        found_object._submit_login()
        return found_object

    def __str__(self):
        return self.key



from django.urls import path, re_path

from avishan.views.class_based import AvishanModelApiView, PasswordHash
from avishan.configure import get_avishan_config
from avishan.views.function_based import avishan_doc, avishan_chayi_create, avishan_redoc

urlpatterns = [
    path(f'{get_avishan_config().AVISHAN_URLS_START}'
         f'/doc',
         avishan_doc),
    path(f'{get_avishan_config().AVISHAN_URLS_START}'
         f'/chayi',
         avishan_chayi_create),
    path(f'{get_avishan_config().AVISHAN_URLS_START}'
         f'/redoc',
         avishan_redoc),
    path(f'{get_avishan_config().AVISHAN_URLS_START}'
         f'/hash_password/<str:password>',
         PasswordHash.as_view()),
    path(f'{get_avishan_config().AVISHAN_URLS_START}'
         f'/<str:model_plural_name>',
         AvishanModelApiView.as_view()),
    path(f'{get_avishan_config().AVISHAN_URLS_START}'
         f'/<str:model_plural_name>/<int:model_item_id>',
         AvishanModelApiView.as_view()),
    path(f'{get_avishan_config().AVISHAN_URLS_START}'
         f'/<str:model_plural_name>/<str:model_function_name>',
         AvishanModelApiView.as_view()),
    path(
        f'{get_avishan_config().AVISHAN_URLS_START}'
        f'/<str:model_plural_name>/<int:model_item_id>/<str:model_function_name>',
        AvishanModelApiView.as_view()
    )
]



# Generated by Django 3.0.3 on 2020-02-14 21:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('avishan', '0003_requesttrackexecinfo'),
    ]

    operations = [
        migrations.AddField(
            model_name='usergroup',
            name='authenticate_with_phone_otp',
            field=models.BooleanField(default=False),
        ),
    ]



import inspect
from datetime import datetime, date, time
from typing import Dict, Callable, Union, Optional
from inspect import Parameter

# todo fake_tree gets list of models needed to be faked. choose relation from themself not independent objects
from django.db import models
from faker import Faker

from avishan.configure import get_avishan_config


class AvishanFaker:
    _faker = Faker(locale=get_avishan_config().FAKER_LOCALE)
    Faker.seed(get_avishan_config().FAKER_SEED)

    @classmethod
    def fake_it(cls, **overriding):
        data = {}
        for key, value in cls.faker_get_create_params_dict().items():

            if callable(getattr(cls, f'_{key}_fake', None)):
                data[key] = getattr(cls, f'_{key}_fake')()
            else:
                if key in overriding.keys():
                    if callable(overriding[key]):
                        faking_method = overriding[key]
                    else:
                        faking_method = None
                else:
                    faking_method = cls.faker_assign_faking_method(key, value)

                if not faking_method:
                    data[key] = overriding[key]
                else:
                    data[key] = faking_method()

        created = cls.faker_get_create_method()(**data)
        if not hasattr(cls, 'fakes'):
            cls.fakes = []
        cls.fakes.append(created)
        return created

    @classmethod
    def fake_list(cls, count: int = 5, **overriding):
        return [cls.fake_it(**overriding) for _ in range(count)]

    @classmethod
    def faker_generate(cls, count: int = 5):
        if not hasattr(cls, 'fakes'):
            cls.fakes = []
        cls.fake_list(count=count - len(cls.fakes))
        return cls.fakes

    @classmethod
    def faker_get_create_params_dict(cls) -> Dict[str, Parameter]:
        return dict(inspect.signature(cls.faker_get_create_method()).parameters.items())

    @classmethod
    def faker_get_create_method(cls) -> Callable:
        from avishan.models import AvishanModel
        cls: Union[AvishanFaker, AvishanModel]
        return cls.create

    @classmethod
    def faker_assign_faking_method(cls, key: str, value: Parameter) -> Callable:
        from avishan.models import AvishanModel

        if not isinstance(value, Parameter):
            raise ValueError(f'parameter {key} in class {cls.__name__} should have type hint to fake it.')
        if key == 'kwargs':
            raise ValueError(f'kwargs in class {cls.__name__} is not valid')

        if value.default is None:
            return lambda: None

        if isinstance(value.annotation, models.base.ModelBase):
            if issubclass(value.annotation, AvishanModel):
                return lambda: value.annotation.fake_it()
            raise NotImplementedError()

        try:
            return {
                bool: cls._bool_fake,
                str: cls._str_fake,
                int: cls._int_fake,
                float: cls._float_fake,
                datetime: cls._datetime_fake,
                date: cls._date_fake,
                time: cls._time_fake
            }[value.annotation]
        except KeyError:
            pass

        if isinstance(value.annotation.__args__[0], ForwardRef):
            list_of = AvishanModel.get_model_with_class_name(value.annotation.__args__[0].__forward_arg__)
        else:
            list_of = value.annotation.__args__[0]
        return lambda: list_of.fake_list()

    @classmethod
    def _bool_fake(cls) -> bool:
        return cls._faker.pybool()

    @classmethod
    def _str_fake(cls) -> str:
        return " ".join(cls._faker.words(8))

    @classmethod
    def _int_fake(cls) -> int:
        return cls._faker.pyint(min_value=0, max_value=100000)

    @classmethod
    def _float_fake(cls,
                    left_digits: Optional[int] = 2,
                    right_digits: Optional[int] = 2,
                    positive: Optional[bool] = True,
                    min_value: Optional[float] = 0,
                    max_value: Optional[float] = None
                    ) -> float:
        return cls._faker.pyfloat(
            left_digits=left_digits,
            right_digits=right_digits,
            positive=positive,
            min_value=min_value,
            max_value=max_value
        )

    @classmethod
    def _datetime_fake(cls) -> datetime:
        return cls._faker.date_time()

    @classmethod
    def _date_fake(cls) -> date:
        return cls._faker.date_object()

    @classmethod
    def _time_fake(cls) -> time:
        return cls._faker.time_object()



from django.http import JsonResponse
from django.shortcuts import render

from avishan.configure import get_avishan_config
from avishan.decorators import AvishanTemplateViewDecorator, AvishanApiViewDecorator


@AvishanTemplateViewDecorator(authenticate=False)
def avishan_doc(request):
    import json
    from avishan.libraries.openapi3.classes import OpenApi
    return render(request, 'avishan/swagger.html',
                  context={'data': json.dumps(OpenApi('0.0.0', 'Documentation').export_json())})


@AvishanApiViewDecorator(authenticate=False)
def avishan_chayi_create(request):
    from avishan.libraries.chayi import ChayiWriter
    ChayiWriter()
    return JsonResponse({'state': 'created'})


@AvishanTemplateViewDecorator(authenticate=False)
def avishan_redoc(request):
    from avishan.libraries.openapi3.rebuild import OpenApi

    open_api_yaml = OpenApi(
        application_title=get_avishan_config().OPENAPI_APPLICATION_TITLE,
        application_description=get_avishan_config().OPENAPI_APPLICATION_DESCRIPTION,
        application_version=get_avishan_config().OPENAPI_APPLICATION_VERSION,
        application_servers=get_avishan_config().OPENAPI_APPLICATION_SERVERS
    ).export_yaml()

    text_file = open('static/openapi.yaml', 'w+')
    text_file.write(open_api_yaml)
    text_file.close()
    return render(request, 'avishan/redoc.html')



from typing import Type, List

import stringcase
from django.db import models
from django.db.models import NOT_PROVIDED

from avishan.models import AvishanModel


def create_dbml_file(address: str):
    dbml = {
        'tables': []
    }
    for model in AvishanModel.get_models():
        dbml['tables'].append(convert_model_to_object(model))
    print(f"DBML READ FOR {len(dbml['tables'])} TABLES")
    from pathlib import Path
    folder = ""
    parts = address.split("/")
    for part in parts[:-1]:
        folder += part + "/"
    Path(folder).mkdir(parents=True, exist_ok=True)
    writer(dbml, address)
    print("DBML FILE CREATED")


def convert_model_to_object(model: Type[AvishanModel]) -> dict:
    data = {
        "name": model.class_name(),
        "columns": [],
        "many_to_many": []
    }
    for field in model.get_full_fields():
        if isinstance(field, models.ManyToManyField):
            data['many_to_many'].append(handle_many_to_many_field(field))
        # todo add note to field
        else:
            data["columns"].append(handle_field(field))
    return data


def handle_field(field: models.Field) -> dict:
    data = {
        'column_name': field.name,
        'column_type': None,
        'column_settings': handle_field_settings(field)
    }
    if isinstance(field, models.OneToOneField):
        data['column_type'] = 'integer'
        data['column_name'] = data['column_name'] + "_id"
        data['column_settings'] = [f"ref: - {field.related_model.class_name()}.id"] + data['column_settings']
    if isinstance(field, models.ForeignKey):
        data['column_type'] = 'integer'
        data['column_name'] = data['column_name'] + "_id"
        data['column_settings'] = [f"ref: > {field.related_model.class_name()}.id"] + data['column_settings']
    elif isinstance(field, (models.BigIntegerField, models.BigAutoField)):
        data['column_type'] = 'long'
    elif isinstance(field, (models.AutoField, models.IntegerField)):
        data['column_type'] = 'integer'
    elif isinstance(field, models.CharField):
        data['column_type'] = f'varchar({field.max_length})'
    elif isinstance(field, models.BooleanField):
        data['column_type'] = 'boolean'
    elif isinstance(field, models.TextField):
        data['column_type'] = 'text'
    elif isinstance(field, models.FloatField):
        data['column_type'] = 'float'
    elif isinstance(field, models.DateTimeField):
        data['column_type'] = 'datetime'
    elif isinstance(field, models.FileField):
        data['column_name'] = 'url'
        data['column_type'] = 'string'
    else:
        raise NotImplementedError()
    if len(data['column_settings']) == 0:
        del data['column_settings']
    return data


def handle_many_to_many_field(field: models.ManyToManyField) -> dict:
    data = {
        "name": ("m2m" + (field.model.class_name()) + stringcase.titlecase(field.name))[:-1],
        "columns": [
            {
                'column_name': stringcase.snakecase(field.model.class_name()) + "_id",
                'column_type': 'integer',
                'column_settings': [f"ref: > {field.model.class_name()}.id"]
            }, {
                'column_name': stringcase.snakecase(field.related_model.class_name()) + "_id",
                'column_type': 'integer',
                'column_settings': [f"ref: > {field.related_model.class_name()}.id"]
            }
        ],
        "many_to_many": []
    }
    data['name'] = data['name'].replace(" ", "")
    return data


def handle_field_settings(field: models.Field) -> list:
    data = []
    if isinstance(field, models.DateTimeField) and (field.auto_now or field.auto_now_add):
        data.append('note: "auto created"')
    if field.primary_key:
        data.append("primary key")
    if field.null:
        data.append("null")
    if field.unique:
        data.append("unique")
    if field.default is not NOT_PROVIDED:
        data.append(f"default: {handle_value(field.default)}")
    return data


def handle_value(value) -> str:
    if value is False:
        return "false"
    if value is True:
        return "true"
    if value is None:
        return "null"
    if isinstance(value, (int, float)):
        return str(value)
    if isinstance(value, str):
        return f'"{value}"'
    raise NotImplementedError()


def writer(dbml: dict, file_address: str):
    lines = []
    for table in dbml['tables']:
        lines += write_table(table)

    f = open(file_address, 'w')
    for line in lines:
        f.write(line + "\n")
    f.close()


def write_table(data: dict) -> List[str]:
    lines = [
        "Table %s {" % data['name']
    ]
    for column in data['columns']:
        lines.append(write_column(column))
    lines.append("}")
    for many in data['many_to_many']:
        lines += write_table(many)
    return lines


def write_column(data) -> str:
    text = f"    {data['column_name']} {data['column_type']}"

    if 'column_settings' in data.keys():
        text += " ["
        for string in data['column_settings']:
            text += string + ", "
        text = text[:-2]
        text += "]"

    return text



import datetime
import typing

from django.db import models

from avishan.configure import get_avishan_config

from inspect import Parameter as InspectParameter

from avishan.models import AvishanModel


class ApiDocumentation:
    paths: typing.List['Path'] = []

    def get_or_create_path(self, url: str) -> 'Path':
        for path in self.paths:
            if path.url == url:
                return path
        return Path(url=url)


# todo example for data
def type_map(input: typing.Union[str, type, models.Field], name: str = None, model=None) -> typing.Tuple[
    typing.Union[str, 'Schema'], typing.Optional[str]]:
    if isinstance(input, str):
        return input, None
    if isinstance(input, Schema):
        return input, None
    if isinstance(input, models.Field):
        if isinstance(input, models.BooleanField):
            return "boolean", None
        if isinstance(input, (models.IntegerField, models.AutoField)):
            return "number", "int64"
        if isinstance(input, models.FloatField):
            return "number", "float"
        if isinstance(input, (models.CharField, models.TextField)):
            return "string", None
        if isinstance(input, models.DateTimeField):
            return "string", "date-time"
        if isinstance(input, models.DateField):
            return "string", "full-date"
        if isinstance(input, models.TimeField):
            return "string", 'full-time'
        if isinstance(input, (models.OneToOneField, models.ForeignKey)):
            if input.related_model.__name__ is model.model.__name__:
                return model, None
            return Schema.create_from_model(input.related_model), None
        if isinstance(input, models.FileField):
            return "string", None
        raise NotImplementedError(name)
    if isinstance(input, type):
        if input is str:
            return 'string', None
        if input is bool:
            return 'boolean', None
        if input is int:
            return 'number', 'int64'
        if input is float:
            return 'number', 'float'
        if input is datetime.datetime:
            return "string", "date-time"
        if input is datetime.date:
            return "string", "full-date"
        if input is datetime.time:
            return "string", "full-time"
        raise NotImplementedError(name)

    # Optional[int] -> Union[int, None]
    # noinspection PyUnresolvedReferences
    if isinstance(input, typing._Union) and len(input.__args__) == 2:
        # noinspection PyUnresolvedReferences
        return type_map(input.__args__[0], name=name)

    raise NotImplementedError(name)


class SchemaProperty:
    def __init__(self, name: str, type: typing.Union[str, type, models.Field, 'Schema'], required: bool = False,
                 raw: bool = False, model=None):
        self.name = name
        self.type, self.format = type_map(type, name, model=model)
        self.required = required
        self.field: typing.Optional[models.Field] = None
        self.raw = raw

    @classmethod
    def create_from_model_field(cls, field: models.Field, model=None) -> 'SchemaProperty':
        schema_property = SchemaProperty(
            name=field.name,
            type=field,
            required=AvishanModel.is_field_required(field),
            model=model
        )
        schema_property.field = field
        return schema_property

    def export_json(self):
        if self.raw:
            return self.type.export_json()
        if not isinstance(self.type, str):
            return {
                '$ref': f"#/components/schemas/{self.type.name}"
            }
        return {
            'type': self.type
        }

    def __str__(self):
        return self.name


class Schema:
    created = {}

    def __init__(self, name: str, type: str = 'object', properties: typing.List[SchemaProperty] = (),
                 items: 'Schema' = None,
                 raw: bool = False):
        self.name = name
        self.type = type
        self.properties = properties
        self.items: typing.Optional[Schema] = items
        self.model: typing.Optional[AvishanModel] = None
        self.raw = raw
        if self.raw:
            for prop in self.properties:
                prop.raw = True

    @classmethod
    def create_from_model(cls, model: typing.Type[AvishanModel]):
        if model not in cls.created.keys():
            schema = Schema(name=model.class_name(), type="object")
            schema.model = model
            schema.properties = schema.create_model_properties(model)
            cls.created[model] = schema
        else:
            schema = cls.created[model]
        return schema

    def create_model_properties(self, model: typing.Type[AvishanModel]) -> typing.List[SchemaProperty]:
        return [SchemaProperty.create_from_model_field(field, model=self) for field in model.get_fields()]

    @classmethod
    def create_from_function(cls, name: str, function):
        import inspect

        properties = []
        for key, value in dict(inspect.signature(function).parameters.items()).items():
            value: InspectParameter
            if key in ['self', 'cls', 'kwargs']:
                continue
            # noinspection PyUnresolvedReferences
            if type(value.annotation) is inspect._empty:
                raise ValueError(f'Method ({function}) parameter ({key}) type not defined')
            if inspect.isclass(value.annotation) and issubclass(value.annotation, AvishanModel):
                param_type = Schema.create_from_model(value.annotation)
            else:
                try:
                    param_type = Schema.create_from_model(value.annotation.__args__[0])
                except:
                    param_type = value.annotation

            properties.append(SchemaProperty(
                name=key,
                type=param_type,
                required=True  # todo check
            ))

        return cls(
            name=name,
            type='object',
            properties=properties
        )

    def export_json(self) -> dict:
        return {
            "type": self.type,
            "properties": self.export_properties_json(),
            "required": self.export_required_json()
        } if self.items is None else {
            'type': self.type,
            'items': self.items.export_json()
        }

    def export_reference(self) -> dict:
        if self.raw:
            return self.export_json()
        return {
            "$ref": f"#/components/schemas/{self.name}"
        } if self.items is None else {
            'type': self.type,
            'items': self.items.export_reference()
        }

    def export_required_json(self) -> typing.List[str]:
        return [item.name for item in self.properties if item.required]

    def export_properties_json(self) -> dict:
        data = {}
        for prop in self.properties:
            data[prop.name] = prop.export_json()
        return data

    @classmethod
    def schema_in_json(cls, name: str, schema: 'Schema'):
        return Schema(
            name=name,
            type='object',
            properties=[SchemaProperty(
                name=name,
                type=schema
            )]
        )

    @classmethod
    def raw_schema_from_dict(cls, data: dict) -> typing.Union['Schema', 'SchemaProperty']:
        import inspect

        name = list(data.keys())[0]
        properties = []
        items = None
        if isinstance(data[name], dict):
            type = 'object'
            for key, value in data[name].items():
                if isinstance(value, dict):
                    inner_type = SchemaProperty(
                        name=key,
                        type=Schema.raw_schema_from_dict(
                            {key: value}
                        )
                    )
                elif isinstance(value, list):
                    inner_type = SchemaProperty(
                        name=key,
                        type=Schema.raw_schema_from_dict({
                            key: [value[0]]
                        })
                    )
                elif inspect.isclass(value) and issubclass(value, AvishanModel):
                    inner_type = SchemaProperty(
                        name=key,
                        type=Schema.create_from_model(
                            value
                        )
                    )
                else:
                    inner_type = SchemaProperty(
                        name=key,
                        type=value
                    )

                properties.append(
                    inner_type
                )
        elif isinstance(data[name], list):
            type = 'array'
            items = Schema(
                name=name,
                type='array',
                items=Schema.raw_schema_from_dict(data[name][0])
            )
        else:
            if issubclass(data[name], AvishanModel):
                return SchemaProperty(
                    name=name,
                    type=Schema.create_from_model(data[name])
                )
            return SchemaProperty(
                name=name,
                type=data[name]
            )
        return Schema(
            name=name,
            type=type,
            properties=properties,
            items=items
        )

    def __str__(self):
        return self.name


class Component:
    def __init__(self, schemas: typing.List[Schema]):
        self.schemas = schemas

    def export_json(self):
        data = {
            'schemas': {}
        }
        for schema in self.schemas:
            data['schemas'][schema.name] = schema.export_json()
        return data


class Content:
    def __init__(self, schema: Schema, type: str, raw: bool = False):
        self.schema = schema
        self.type = type
        self.raw = raw

    def export_json(self) -> dict:
        return {
            'schema': self.schema.export_reference()
        } if not self.raw else {
            'schema': self.schema.export_json()
        }


def __str__(self):
    return f'{self.schema}'


class ContentArray(Content):

    def __init__(self, schema: Schema, type: str):
        super().__init__(schema, type)


class Parameter:
    def __init__(self, name: str, schema: Schema, where: str = 'query', description: str = None, required: bool = False,
                 style: str = None):
        self.name = name
        self.schema = schema
        self.where = where
        self.description = description
        self.required = required
        self.style = style

    def export_json(self):
        pass  # todo

    def __str__(self):
        return self.name


class PathRequest:
    def __init__(self, required: bool = True, contents: typing.List[Content] = ()):
        self.required = required
        self.contents = contents

    def export_json(self) -> dict:
        data = {
            'required': self.required,
        }
        if len(self.contents) > 0:
            data['content'] = {}
        for content in self.contents:
            data['content'][content.type] = content.export_json()
        return data

    def __str__(self):
        return f'{len(self.contents)} requests'


class PathResponse:
    def __init__(self, status_code: int, contents: typing.List[Content], description: str = None):
        self.status_code = status_code
        self.contents = contents
        self.description = description

    def export_json(self) -> dict:
        data = {}
        if len(self.contents) > 0:
            data['content'] = {}
        for content in self.contents:
            data['content'][content.type] = content.export_json()
        if self.description:
            data['description'] = self.description
        return data

    def __str__(self):
        return self.status_code


class PathResponseGroup:
    def __init__(self, responses: typing.List[PathResponse]):
        self.responses = responses

    def export_json(self) -> dict:
        data = {}
        for item in self.responses:
            data[item.status_code] = item.export_json()
        return data

    def __str__(self):
        return f'{len(self.responses)} responses'


class PathMethod:
    method = None

    def __init__(self, summary: str = None, description: str = None, request: PathRequest = None,
                 responses: PathResponseGroup = None, tags: typing.List[str] = ()):
        self.request = request
        if responses is not None and len(responses.responses) == 0:
            self.responses = None
        else:
            self.responses = responses

        self.tags = tags
        self.summary = summary
        self.description = description

    def export_json(self) -> dict:
        # todo added parameters
        data = {}
        if self.request:
            data['requestBody'] = self.request.export_json()
        if self.responses:
            data['responses'] = {}
            for response in self.responses.responses:
                data['responses'][response.status_code] = response.export_json()
        if len(self.tags) > 0:
            data['tags'] = self.tags
        if self.summary:
            data['summary'] = self.summary
        if self.description:
            data['description'] = self.description
        return data

    def __str__(self):
        return f'{self.method.upper()}'


class PathGetMethod(PathMethod):
    method = 'get'

    def __init__(self, summary: str = None, description: str = None,
                 responses: PathResponseGroup = None, tags: typing.List[str] = ()):
        super().__init__(summary, description, None, responses, tags)


class PathPostMethod(PathMethod):
    method = 'post'


class PathPutMethod(PathMethod):
    method = 'put'


class PathDeleteMethod(PathMethod):
    method = 'delete'

    def __init__(self, summary: str = None, description: str = None,
                 responses: PathResponseGroup = None, tags: typing.List[str] = ()):
        super().__init__(summary, description, None, responses, tags)


class Path:
    def __init__(self, url: str, methods: typing.List[PathMethod] = (), description: str = "", parameters: list = None):
        self.methods = methods
        self.url = url
        self.description = description
        self.parameters = OpenApi.request_common_url_parameters()
        if parameters:
            parameters.extend(parameters)

    def export_json(self):
        data = {}
        for path_method in self.methods:
            data[path_method.method] = path_method.export_json()

        return data

    def __str__(self):
        return self.url


class OpenApi:
    def __init__(self, api_version: str, api_title: str, open_api_version: str = "3.0.0", api_description: str = None,
                 servers: typing.Tuple[str] = ()):
        self.api_version = api_version
        self.api_title = api_title
        self.api_description = api_description
        self.servers = servers
        self.open_api_version = open_api_version
        self.models: typing.List[typing.Type[AvishanModel]] = []
        self.schemas = self.create_schemas_from_models()
        self.paths = self.create_paths_from_views()

        # todo security

    @staticmethod
    def request_common_url_parameters() -> typing.List[dict]:
        # todo load from dict
        return get_avishan_config().REQUEST_COMMON_URL_PARAMETERS

    def create_schemas_from_models(self) -> typing.List['Schema']:
        def all_subclasses(cls):
            return set(cls.__subclasses__()).union(
                [s for c in cls.__subclasses__() for s in all_subclasses(c)])

        schemas = []
        for model in sorted(all_subclasses(AvishanModel), key=lambda x: x.class_name()):
            model: typing.Type[AvishanModel]
            self.models.append(model)
            schemas.append(Schema.create_from_model(model))
        return schemas

    @staticmethod
    def create_paths_from_views() -> typing.List['Path']:
        from avishan.views.class_based import AvishanModelApiView

        if AvishanModelApiView.documentation():
            return AvishanModelApiView.documentation().paths
        else:
            return []

    def export_schemas_json(self) -> dict:
        data = {}
        for schema in self.schemas:
            data[schema.name] = schema.export_json()
        return data

    def export_paths_json(self) -> dict:
        data = {}
        for path in self.paths:
            data[path.url] = path.export_json()
        return data

    def export_json(self) -> dict:
        # todo tags
        data = {
            'openapi': self.open_api_version,
            'info': {
                'version': self.api_version,
                'title': self.api_title
            },
            'components': {
                'schemas': self.export_schemas_json()
            },
            'paths': self.export_paths_json()
        }
        if self.api_description:
            data['info']['description'] = self.api_description
        if len(self.servers) > 0:
            data['servers'] = [{'url': item} for item in self.servers]
        return data



import inspect
from inspect import Parameter
from typing import Type, List, Tuple, Dict

from django.db import models

from avishan.models import AvishanModel
import stringcase


def request_common_parameters() -> List[dict]:
    return [{
        "name": 'lang',
        "in": 'query',
        "description": 'set language for this request',
        "required": False,
    }]


def create_openapi_object(title: str, api_version: str, servers: List[dict] = None) -> dict:
    data = {
        'openapi': "3.0.0",
        'info': get_info_object(api_version, title),
        'servers': servers,
        'components': {
            'schemas': {}
        },
        'paths': {}
    }
    for model in AvishanModel.get_models():
        data['components']['schemas'][model.class_name()] = (get_model_object(model))
        data['paths'] = {**data['paths'], **get_model_paths(model)}
    return data


def get_info_object(api_version: str, title: str):
    # todo
    return {
        'version': api_version,
        'title': title
    }


def get_model_paths(model: Type[AvishanModel]) -> dict:
    data = {}
    data[f'/api/av1/{model.class_plural_snake_case_name()}'] = {
        "get": get_model_get_request(model),
        "post": get_model_post_request(model)
    }
    data[f'/api/av1/{model.class_plural_snake_case_name()}' + '/{item_id}'] = {
        "get": get_item_get_request(model),
        "put": get_item_put_request(model),
        "delete": get_item_delete_request(model),
    }
    for key, value in data.items():
        deletes = []

        for inner_key, inner_value in data[key].items():
            if inner_value == {}:
                deletes.append(inner_key)
        for item in deletes:
            del data[key][item]
    return data


def get_model_get_request(model: Type[AvishanModel]):
    return {
        "description": f"Get list of {stringcase.titlecase(model.class_name())}",
        "parameters": request_common_parameters(),
        "responses": {
            "200": {
                "content": {
                    "application/json": {
                        "schema": {
                            "type": "array",
                            "items": {
                                "$ref": f"#/components/schemas/{model.class_name()}"
                            }
                        }
                    }
                },
                "description": "succeed"
            }
        }
    }


def get_model_post_request(model: Type[AvishanModel]) -> dict:
    if model.Meta.abstract is True:
        return {}
    create_schema = get_model_create_schema(model)
    if create_schema == {}:
        return {}
    return {
        "description": f"Create {stringcase.titlecase(model.class_name())}",
        "parameters": request_common_parameters(),
        "requestBody": {
            "description": "Data structure for creation",
            "content": {
                "application/json": {
                    "schema": create_schema
                }
            },
            "required": True
        },
        "responses": {
            "200": {
                "content": {
                    "application/json": {
                        "schema": {
                            "type": "array",
                            "items": {
                                "$ref": f"#/components/schemas/{model.class_name()}"
                            }
                        }
                    }
                },
                "description": "succeed"
            }
        }
    }


def get_item_get_request(model: Type[AvishanModel]) -> dict:
    return {
        "description": f"Get item of {stringcase.titlecase(model.class_name())}",
        "parameters": request_common_parameters(),
        "responses": {
            "200": {
                "content": {
                    "application/json": {
                        "schema": {
                            "type": "object",
                            "properties": {
                                model.class_snake_case_name(): {
                                    "$ref": f"#/components/schemas/{model.class_name()}"
                                }
                            }
                        }
                    }
                },
                "description": "succeed"
            }
        }
    }


def get_item_put_request(model: Type[AvishanModel]) -> dict:
    if model.Meta.abstract is True:
        return {}
    update_schema = get_model_update_schema(model)
    if update_schema == {}:
        return {}
    return {
        "description": f"Update {stringcase.titlecase(model.class_name())} item",
        "parameters": request_common_parameters(),
        "requestBody": {
            "description": "Data structure for edit",
            "content": {
                "application/json": {
                    "schema": update_schema
                }
            },
            "required": True
        },
        "responses": {
            "200": {
                "content": {
                    "application/json": {
                        "schema": {
                            "type": "object",
                            "properties": {
                                model.class_snake_case_name(): {
                                    "$ref": f"#/components/schemas/{model.class_name()}"
                                }
                            }
                        }
                    }
                },
                "description": "succeed"
            }
        }
    }


def get_item_delete_request(model: Type[AvishanModel]) -> dict:
    if model.Meta.abstract is True:
        return {}
    return {
        "description": f"Delete item_id from {stringcase.titlecase(model.class_name())}",
        "parameters": request_common_parameters(),
        "responses": {
            "200": {
                "content": {
                    "application/json": {
                        "schema": {
                            "type": "object",
                            "properties": {
                                model.class_snake_case_name(): {
                                    "$ref": f"#/components/schemas/{model.class_name()}"
                                }
                            }
                        }
                    }
                },
                "description": "succeed"
            }
        }
    }


def get_functions_properties(function) -> Dict[str, Parameter]:
    return dict(inspect.signature(function).parameters.items())


def get_model_create_schema(model: Type[AvishanModel]) -> dict:
    data = {
        "type": "object",
        "properties": {}
    }
    create_function = getattr(model, 'create')
    for name, param in get_functions_properties(create_function):
        if name == 'kwargs':
            return {}
        data['properties'][name] = get_input_object(param)
    return data


def get_model_update_schema(model):
    import inspect
    data = {
        "type": "object",
        "properties": {}
    }
    update_function = getattr(model, 'update')

    update_signature = inspect.signature(update_function)
    for name, param in update_signature.parameters.items():
        if name == 'self':
            continue
        if name == 'kwargs':
            return {}
        data['properties'][name] = get_input_object(param)
    return data


def get_input_object(param: Parameter) -> dict:
    data = {
        'type': param.annotation,
        'default': param.default
    }
    if param.annotation is _empty:
        raise ValueError(f"Documentation Parse failed: {param}")
    if param.default is _empty:
        del data['default']

    if data['type'] == int:
        data['type'] = "integer"
    elif data['type'] == float:
        data['type'] = "float"
    elif data['type'] == str:
        data['type'] = "string"
    elif data['type'] == bool:
        data['type'] = 'boolean'
    elif inspect.isclass(data['type']) and issubclass(data['type'], AvishanModel):
        data = {
            "$ref": f"#/components/schemas/{data['type'].class_name()}"
        }
        return data
    else:
        try:
            data = {
                'type': "array",
                "items": {
                    "$ref": f"#/components/schemas/{data['type'].__args__[0].class_name()}"
                }
            }
        except:
            raise NotImplementedError()
    return data


def get_model_object(model: Type[AvishanModel]) -> dict:
    return {
        "type": "object",
        "properties": get_model_properties(model),
        "required": get_model_required_fields(model)
    }  # todo abstract classes


def get_model_properties(model: Type[AvishanModel]) -> dict:
    data = {}
    for field in model.get_fields():
        data = {**data, **get_field_object(field)}

    return data


def get_field_object(field: models.Field) -> dict:
    data = {
        field.name: None
    }
    if isinstance(field, models.BooleanField):
        data[field.name] = {
            'type': "boolean"
        }
    elif isinstance(field, (models.IntegerField, models.AutoField)):
        data[field.name] = {
            'type': "number"
        }
    elif isinstance(field, models.FloatField):
        data[field.name] = {
            'type': "float"
        }
    elif isinstance(field, (models.CharField, models.TextField)):
        data[field.name] = {
            'type': "string"
        }
    elif isinstance(field, models.DateTimeField):
        data[field.name] = {
            'type': "date-time"
        }
    elif isinstance(field, models.DateField):
        data[field.name] = {
            'type': "date"
        }
    elif isinstance(field, models.TimeField):
        data[field.name] = {
            'type': "string"
        }
    elif isinstance(field, (models.OneToOneField, models.ForeignKey)):
        data[field.name] = {
            '$ref': f"#/components/schemas/{field.related_model.class_name()}"
        }
    elif isinstance(field, models.FileField):
        data[field.name] = {
            'type': 'object',
            'properties': {
                'id': {
                    'type': 'integer'
                },
                'url': {
                    'type': 'string'
                }
            }
        }
    else:
        raise NotImplementedError()
    if len(field.choices) > 0:
        data[field.name]['enum'] = [item[1] for item in field.choices]
    return data


def get_field_added_data(field: models.Field) -> dict:
    data = {}
    if field.null:
        data['nullable']: True
    return data


def get_model_required_fields(model: Type[AvishanModel]) -> list:
    data = []
    for field in model.get_full_fields():
        if AvishanModel.is_field_required(field):
            data.append(field.name)

    return data



# Generated by Django 3.0.8 on 2020-07-26 11:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('avishan', '0015_auto_20200724_1419'),
    ]

    operations = [
        migrations.AlterField(
            model_name='country',
            name='flag_url',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
    ]



import datetime
from typing import Optional, Union, List

from django.http import HttpResponse
from django.utils import timezone

from avishan.exceptions import AuthException, AvishanException
from . import current_request
from .configure import get_avishan_config
from .misc import status
from .models import AuthenticationType


class AvishanDataValidator:
    class ValidatorException(AvishanException):
        from avishan.misc.translation import AvishanTranslatable

        def __init__(self, field_name: Union[str, 'AvishanTranslatable']):
            from avishan.misc.translation import AvishanTranslatable
            current_request['response']['error_in_field'] = field_name
            current_request['response']['error_message'] = AvishanTranslatable(
                EN=f'"{field_name}" not accepted',
                FA='"' + field_name + '" قابل قبول نیست'
            )
            current_request['status_code'] = status.HTTP_417_EXPECTATION_FAILED
            super().__init__()

    @classmethod
    def validate_phone_number(cls, input: str, country_code: str = '98', phone_start_number: str = '09') -> str:
        from avishan.misc.translation import AvishanTranslatable

        input = en_numbers(input)
        input = input.replace(" ", "")
        input = input.replace("-", "")

        if input.startswith("00"):
            if not input.startswith("00" + country_code):
                raise cls.ValidatorException(AvishanTranslatable(
                    EN='phone number', FA='شماره موبایل'
                ))
            if input.startswith("00" + country_code + phone_start_number):
                input = "00" + country_code + input[5:]
        elif input.startswith("+"):
            if not input.startswith("+" + country_code):
                raise cls.ValidatorException(AvishanTranslatable(
                    EN='phone number', FA='شماره موبایل'
                ))
            input = "00" + input[1:]
            if input.startswith("00" + country_code + phone_start_number):
                input = "00" + country_code + input[5:]
        elif input.startswith(phone_start_number):
            input = "00" + country_code + input[1:]

        if len(input) != 14 or not input.isdigit():
            raise cls.ValidatorException(AvishanTranslatable(
                EN='phone number', FA='شماره موبایل'
            ))

        return input

    @classmethod
    def validate_text(cls, input: str, blank: bool = True) -> str:
        input = input.strip()
        input = fa_numbers(input)

        if not blank and len(input) == 0:
            from avishan.misc.translation import AvishanTranslatable
            raise cls.ValidatorException(AvishanTranslatable(EN='text', FA='متن'))

        return input

    @classmethod
    def validate_recommend_code(cls, input: str) -> str:
        input = cls.validate_text(input)

        input = en_numbers(input)

        input = input.upper()

        return input

    @classmethod
    def validate_first_name(cls, input):
        input = input.strip()

        if has_numbers(input) or len(input) < 2:
            from avishan.misc.translation import AvishanTranslatable
            raise cls.ValidatorException(AvishanTranslatable(EN='first name', FA='نام'))

        return input

    @classmethod
    def validate_last_name(cls, input):
        input = input.strip()

        if has_numbers(input) or len(input) < 2:
            from avishan.misc.translation import AvishanTranslatable
            raise cls.ValidatorException(AvishanTranslatable(EN='last name', FA='نام خانوادگی'))

        return input

    @classmethod
    def validate_ferdowsi_student_id(cls, input):
        input = cls.validate_text(input, blank=False)

        if not input.isdigit():
            from avishan.misc.translation import AvishanTranslatable
            raise cls.ValidatorException(AvishanTranslatable(EN='Student ID', FA='کد دانشجویی'))
        return input

    @classmethod
    def validate_plate(cls, plate_a, plate_b, plate_c, plate_d):
        from avishan.misc.translation import AvishanTranslatable
        plate_a = cls.validate_text(fa_numbers(plate_a), blank=False)
        plate_b = cls.validate_text(fa_numbers(plate_b), blank=False)
        plate_c = cls.validate_text(fa_numbers(plate_c), blank=False)
        plate_d = cls.validate_text(fa_numbers(plate_d), blank=False)

        if plate_b not in ['ب', 'ج', 'د', 'س', 'ص', 'ط', 'ق', 'ل', 'م', 'ن', 'و', 'ه', 'ی', 'الف', 'پ', 'ت', 'ث', 'ز',
                           'ژ',
                           'ش', 'ع', 'ف', 'ک', 'گ', 'D', 'S', 'd', 's', 'ي']:
            raise cls.ValidatorException(AvishanTranslatable(EN='plate', FA='پلاک'))

        if not plate_a.isdigit() or not plate_c.isdigit() or not plate_d.isdigit():
            raise cls.ValidatorException(AvishanTranslatable(EN='plate', FA='پلاک'))

        return plate_a, plate_b, plate_c, plate_d

    @classmethod
    def validate_time(cls, input: dict, name: str) -> datetime.time:
        return datetime.time(int(input['hour']), int(input['minute']))


def discard_monitor(url: str) -> bool:
    """
    checks if request is in check-blacklist
    :param url: request url. If straightly catch from request.path, it comes like: /admin, /api/v1
    :return:
    """
    if url.startswith(tuple(get_avishan_config().NOT_MONITORED_STARTS)):
        return True
    return False


def find_token_in_header() -> bool:
    """
    find token and put it in current_request
    :return: false if token not found
    """
    try:
        temp = current_request['request'].META['HTTP_AUTHORIZATION']
        if can_be_token(temp[6:]):
            current_request['token'] = temp[6:]
            return True
    except KeyError:
        pass

    return False


def find_token_in_session() -> bool:
    """
    find token and put it in current_request
    :return: false if token not found
    """
    try:
        temp = current_request['request'].COOKIES['token']
        if can_be_token(temp):
            current_request['token'] = temp
            return True
    except KeyError:
        pass

    return False


def find_token() -> bool:
    """
    check for token in both session and header
    :return: true if token
    """
    if current_request['request'].path.startswith('/api/v1/login/generate/'):
        return False
    if not find_token_in_header() and not find_token_in_session():
        return False
    return True


def can_be_token(text):
    """
    checks entered text can be a token
    :param text:
    :return:
    """
    # todo 0.2.2: can we use jwt package itself for check?
    if len(text) > 0:
        return True
    return False


def add_token_to_response(rendered_response: HttpResponse):
    """
    create new token if needed, else reuse previous
    add token to session if session-based auth, else to response header
    """
    if current_request['json_unsafe']:
        return
    if not current_request['add_token'] or not current_request['authentication_object']:
        delete_token_from_request(rendered_response)
    else:
        token = encode_token(current_request['authentication_object'])

        if current_request['is_api']:
            current_request['response']['token'] = token
        else:
            rendered_response.set_cookie('token', token)


def delete_token_from_request(rendered_response=None):
    if current_request['is_api']:
        try:
            del current_request['response']['token']
        except KeyError:
            pass
    else:
        rendered_response.delete_cookie('token')


def encode_token(authentication_object: 'AuthenticationType') -> Optional[str]:
    import jwt
    from datetime import timedelta

    now = timezone.now()
    token_data = {
        'at_n': authentication_object.class_name(),
        'at_id': authentication_object.id,
        'exp': (now + timedelta(
            seconds=authentication_object.user_user_group.user_group.token_valid_seconds)).timestamp(),
        'crt': now.timestamp(),
        'lgn': authentication_object.last_login.timestamp() if authentication_object.last_login else now.timestamp()
    }
    return jwt.encode(token_data,
                      get_avishan_config().JWT_KEY,
                      algorithm='HS256'
                      ).decode("utf8")


def decode_token():
    import jwt
    if not current_request['token']:
        raise AuthException(AuthException.TOKEN_NOT_FOUND)
    try:
        current_request['decoded_token'] = jwt.decode(
            current_request['token'], get_avishan_config().JWT_KEY,
            algorithms=['HS256']
        )
        current_request['add_token'] = True
    except jwt.exceptions.ExpiredSignatureError:
        raise AuthException(AuthException.TOKEN_EXPIRED)
    except:
        raise AuthException(AuthException.ERROR_IN_TOKEN)


def find_and_check_user():
    """
    Populate current_request object with data from token. Then check for user "active" authorization
    :return:
    """
    from avishan.models import AvishanModel

    if not current_request['decoded_token']:
        AuthException(AuthException.ERROR_IN_TOKEN)

    authentication_type_class = AvishanModel.get_model_with_class_name(
        current_request['decoded_token']['at_n']
    )
    try:
        authentication_type_object: 'AuthenticationType' = authentication_type_class.objects.get(
            id=current_request['decoded_token']['at_id'])
        user_user_group = authentication_type_object.user_user_group
    except authentication_type_class.DoesNotExist:
        raise AuthException(AuthException.ACCOUNT_NOT_FOUND)

    if not user_user_group.is_active:
        raise AuthException(AuthException.GROUP_ACCOUNT_NOT_ACTIVE)
    if not user_user_group.base_user.is_active:
        raise AuthException(AuthException.ACCOUNT_NOT_ACTIVE)
    if authentication_type_object.last_login is None or \
            authentication_type_object.last_login.timestamp() != current_request['decoded_token']['lgn'] \
            or authentication_type_object.last_logout:
        raise AuthException(AuthException.DEACTIVATED_TOKEN)

    authentication_type_object.last_used = timezone.now()
    authentication_type_object.save()
    authentication_type_object._populate_current_request()


def create_avishan_config_file(app_name: str = None):
    # todo 0.2.0 create config file and its classes. add needed fields
    """
    MONITORED_APPS_NAMES = []
    NOT_MONITORED_STARTS [
        '/admin', '/static', '/media', '/favicon.ico'
    ]
    JWT_KEY = "" or none if not available
    """
    if app_name:
        f = open(app_name + "/avishan_config.py", 'w+')
    else:
        f = open('avishan_config.py', 'w+')
    f.writelines((
        'def check():\n',
        '    pass\n\n\n',
        'class AvishanConfig:\n'
    ))
    if not app_name:
        f.writelines([
            '    MONITORED_APPS_NAMES = []\n',
            "    NOT_MONITORED_STARTS = ['/admin', '/static', '/favicon.ico']\n",
            "    JWT_KEY = 'CHANGE_THIS_KEY'\n",
            "    USE_JALALI_DATETIME = True\n",
        ])
    else:
        f.write("    pass\n")
    f.close()


def fa_numbers(text):
    text = str(text)
    text = en_numbers(text)
    array = ['۰', '۱', '۲', '۳', '۴', '۵', '۶', '۷', '۸', '۹']
    result = ''
    for i in str(text):
        if i.isdigit():
            result = result + array[int(i)]
        else:
            result = result + i

    return result


def en_numbers(text):
    text = str(text)
    result = ''
    array = ['۰', '۱', '۲', '۳', '۴', '۵', '۶', '۷', '۸', '۹']
    for char in text:
        if char in array:
            result += str(array.index(char))
        else:
            result += char
    return result


def has_numbers(input):
    return any(char.isdigit() for char in input)


def find_file(name: str, parent_directory_path: str) -> List[str]:
    import os

    result = []
    for root, dirs, files in os.walk(parent_directory_path):
        if name in files:
            result.append(os.path.join(root, name))
    return result


def all_subclasses(parent_class):
    return list(set(parent_class.__subclasses__()).union(
        [s for c in parent_class.__subclasses__() for s in all_subclasses(c)]))


def parse_url(url: str):
    from urllib.parse import urlparse
    return urlparse(url)
